% domino.pl llomny
% A domin feladvny megoldsa irny- s domin-vltozkkal
% Nagyhatkonysg Logikai Programozs 2000. tavaszi flvi hzi feladat
% 1.4 Verzi, 2000. aug. 4.
% (c) Szeredi Pter

:- module(domino_szp1, [domino/2, domino/3, option_variant/2]).

:- use_module(library(lists)).
:- use_module(library(clpfd)).
:- use_module('../utils', [append_lists/2, transpose/2,
			   same_size_matrix/2]).
:- use_module('../testenv', [filter_vars/3]).

% A megvalsts alapelvei
% ------------------------

% Korlt vltozk:

% a. A domin-tbla minden elemnek megfeleltetnk egy korlt-vltozt,
%    a ngy gtjnak megfelel kdrtkkel. Ezeket az domin-tblnak
%    megfelelen mtrixba rendezzk (ez az irnymtrix, elemei az
%    irny-vltozk). 
%
% b. Minden dominnak megfeleltetnk egy korlt-vltozt (ezt 
%    domin-vltoznak hvjuk).  Ha a domint n-flekppen lehet letenni a
%    tblra, akkor a vltoz tartomnya 1..n lesz. A domin-vltoz rtke
%    csak azt mondja meg, hogy a lehetsges n lehelyezs kzl ez a
%    hanyadik. A domin leraks koordintit ez teht nem hatrozza meg, de
%    erre a megoldsban nincs is szksg. 

% Korltok:

% Szomszdsgi korltok:
%    az irnymtrix szak-dli s nyugat-keleti irnyban szomszdos mezire
%    fellltjuk a megfelel korltokat (pl. a fels szomszd akkor s csak
%    akkor lehet szak, ha az als dl).
% Domin korltok:
%    minden domin s annak minden leraksa esetre fellltunk egy
%    korltot, amely a domin-vltozt sszekapcsolja az adott leraks els
%    (azaz baloldali vagy fels) mezjnek megfelel irny-vltozval. 
%    Pldul, ha egy domin 2. leraksa fggleges, s a fels mez a 3. sor
%    4. oszlopba esik, akkor az ennek a leraksnak megfelel korlt: a 
%    domin-vltoz akkor s csak akkor 2 rtk, ha az irnymtrix
%    3. sornak 4. eleme az szaknak megfelel kd.

% Mindkt korltfajta azonos fajtj, akkor-s-csak-akkor (csakkor)
% kapcsolatot r le:
%                X #= C #<=> Y #= D
% ahol C s D konstansok. A reifikcin alapul nyilvnval megolds
% mellett kt FD-prediktumos megvalstst is tartalmaz a program. Ezek
% kzl a megfelel opcival vlaszthatjuk ki az adott futs sorn
% alkalmazand vltozatot.

% Cmkzs:

% Opci-megadssal vlaszthatunk az irny- vagy a domin-vltozk cmkzse
% kztt.

% A cmkzs sorn fontos szerepet kap az n. szrs. Szrskor adott egy
% vltoz-lista s rtkek egy listja. Sorra vesszk a vltozkat s
% mindegyiket megprbljuk mindegyik rtkre behelyettesteni. Ha egy
% behelyettests meghisulst okoz (kzvetlenl, tovbbi cmkzs nlkl),
% akkor az adott rtket kivesszk a vltoz tartomnybl, egybknt nem
% csinlunk semmit. Fontos megjegyezni, hogy a szrs nem hoz ltre
% vlasztsi pontot.

% Ebben a programban mindenkppen az irny-vltozkat szrjk. A szrsi
% rtk-lista alaphelyzetben az szak s nyugat kdjt tartalmazza, de ez
% opcival fellbrlhat.  Alaprtelmezsben nem szrnk, de opcival
% elrhatjuk, hogy szrjnk, s azt is, hogy milyen gyakorisggal
% (meghatrozhatjuk, hogy a labeling sorn csak minden N. vltoz esetn
% szrjnk).

% A tnyleges cmkzs teht a kivlasztott vltoz-listn, a kivlasztott
% szrsi mdszerrel trtnik. Opcival adhatjuk meg a labeling/2 eljrs
% cmkzsi-opci-listjt is.

% A program vzlata:

% Felptjk a domin-tblval azonos mret irnymtrixot, felvesszk a
% szomszdsgi korltokat. Kigyjtjk az egyes domink leraksi
% lehetsgeit (lsd a `lerakasok' tpust albb), s minden
% domin-leraksra  felvesszk a domin korltokat.

% Vgl elvgezzk a cmkzst (belertve a szrst), majd a kapott
% megolds-mtrixban az irnykdokat irny-atomokra cserljk.

% Adattpusok
% -----------

% :- type lerakasok == list(domino_lerakasok).
% :- type domino_lerakasok == list(pozirany).
% :- type pozirany ---> pozicio-iranykod.
% :- type pozicio ---> sori*oszlopi.         % 1-tol szmozva
% :- type sori == int.
% :- type oszlopi == int.
% :- type iranykod == int.                   % 1 2 3 4
% :- type domino ---> pottyok:pottyok
% :- type pottyok == int.                    % 0 ...
% :- type domino_lerakasi_sorszam == int.
% :- type iranykod_matrix == matrix(iranykod).
% :- type irany_matrix == matrix(irany).
% :- type irany ---> w ; n ; e ; s.
% :- type domino_matrix == matrix(pottyok).

% :- type opciok == list(any).

% :- type matrix(Type) == list(list(Type)).

% Algoritmus-vltozatok
% ---------------------

% A domino/3 belpsi pont harmadik argumentuma segtsgvel az
% algoritmus klnfle vltozatait prblhatjuk ki.

% Az opcilista lehetsges elemei

% csakkor=ind1 (alaprt.)	a csakkor prediktumot egy egyszer
% 				indexiklissal valstjuk meg.
% csakkor=ind2 			a csakkor prediktumot egy sszetettebb
% 				indexiklissal valstjuk meg.
% csakkor=reif			a csakkor prediktumot reifikcival
% 				valstjuk meg.
%

% :- type csvalt ---> ind1 ; ind2 ; reif.

% valt=irany (alaprt.)		Az irny-vltozkat cmkzzk.
% valt=domino 			A domin-vltozkat cmkzzk.

% % :- type mit_cimkez ---> irany ; domino.

% szur=ki (alaprt.)		Nincs szrs.
% szur=elott                	Csak a cmkzs eltt szrnk.
% szur=N	        	A cmkzs eltt szrnk, majd minden N. 
% 				vltoz cmkzse utn ismt.

% szurtek=Ek (alaprt. [1,2])	Szrskor az Ek egsz-lista elemeit prbljuk
% 				behelyettesteni az irny-vltozkba.

% label=LOpciok			LOpciok megadja a labeling/2 eljrsnak 
%  (alaprt. [])		tovbbadand opcikat.

% :- type labeling_opciok == list(any).

% Alaprtelmezsek

% Az opcik alaprtelmezst gy vlasztottuk meg, hogy azok egy viszonylag
% egyszer megoldsnak feleljenek meg. A domino(Mx, IMx, []) hvs teht
% ezt az egyszer megoldst futtatja. A domino(Mx, IMx) hvst viszont nem
% az elzre vezetjk vissza, hanem egy olyan opci-kszletet hasznlunk,
% amely a nehz feladatok megoldsnl hasznos.

% opcio_alap(Nev, Ertek): A Nev nev opci alaprtelmezse Ertek.
% :- pred opcio_alap(atom::in, any::out).
opcio_alap(csakkor, ind1).
opcio_alap(valt,    irany).
opcio_alap(szur,    ki).
opcio_alap(szurtek, [1,2]).
opcio_alap(label,   []).

% eros_opciok(Opciok): Opciok a domino/2-ben hasznland opcilista.
% :- pred eros_opciok(opciok::out).
eros_opciok([valt=domino,szur=2,label=[ff]]).

% Nhny opcilista-varinst kivlasztottunk s nvvel lttunk el. Ennek
% szerepe ketts: a varinsnv megadhat a domino/3 harmadik
% argumentumaknt. Ezen fell a keretprogram, ha nem egy konkrt varins
% futtatsra krik fel, az sszes megnevezett varinst futtatja (ezeket az
% option_variant/2 exportlt prediktum sorolja fel).

% :- type varians == any.  % opcilista vagy atom

% Varians egy opci-varins neve.
% :- pred option_variant(atom::out).
option_variant(Variant, Level) :-
	opcio_varians(Variant, Level, _).

% :- type szint ---> easy ; diff ; hard.

% opcio_varians(Varians, Szint, Opciok): A Varians varinshoz tartoz 
% opci-lista Opciok. Ezt a varinst legfeljebb a Szint szint feladatokra
% rdemes futtatni.
% :- pred opcio_varians(atom::out, szint::out, opciok::out).
opcio_varians(irany_1,      diff, [valt=irany,szur=1,label=[]]).
opcio_varians(irany_2,      hard, [valt=irany,szur=2,label=[]]).
opcio_varians(irany_3,      hard, [valt=irany,szur=3,label=[]]).
opcio_varians(irany_4,      hard, [valt=irany,szur=4,label=[]]).
opcio_varians(irany_5,      hard, [valt=irany,szur=5,label=[]]).
				  			
opcio_varians(irany_2_ffc,  diff, [valt=irany,szur=2,label=[ffc]]).
opcio_varians(irany_2_ff,   diff, [valt=irany,szur=2,label=[ff]]).
opcio_varians(irany_2_ind2, diff, [csakkor=ind2,valt=irany,szur=2,label=[]]).
opcio_varians(irany_2_reif, diff, [csakkor=reif,valt=irany,szur=2,label=[]]).
				  
opcio_varians(irany_2_sz1,  diff, [valt=irany,szur=2,label=[],szurtek=[1]]).
opcio_varians(irany_elott,  easy, [valt=irany,szur=elott,label=[]]).
opcio_varians(irany_ki,     easy, [valt=irany,szur=ki,label=[]]).

opcio_varians(domino_1_ff,  diff, [valt=domino,szur=1,label=[ff]]).
opcio_varians(domino_2_ff,  hard, [valt=domino,szur=2,label=[ff]]).
opcio_varians(domino_3_ff,  hard, [valt=domino,szur=3,label=[ff]]).
opcio_varians(domino_4_ff,  hard, [valt=domino,szur=4,label=[ff]]).
opcio_varians(domino_5_ff,  hard, [valt=domino,szur=5,label=[ff]]).
				  			
opcio_varians(domino_2_ffc, diff, [valt=domino,szur=2,label=[ffc]]).
opcio_varians(domino_2,     hard, [valt=domino,szur=2,label=[]]).
opcio_varians(domino_2_ind2,diff, [csakkor=ind2,valt=domino,szur=2,label=[]]).
opcio_varians(domino_2_reif,diff, [csakkor=reif,valt=domino,szur=2,label=[]]).
				  
opcio_varians(domino_2_sz1, diff, [valt=domino,szur=2,label=[],szurtek=[1]]).
opcio_varians(domino_elott, easy, [valt=domino,szur=elott,label=[]]).
opcio_varians(domino_ki,    easy, [valt=domino,szur=ki,label=[]]).

				  			
% Fbejratok
% -----------

:- public domino/2, domino/3.

% Az Mx domin-mtrix ltal adott domin feladat megoldsa az IMx
% irnymtrix.
% :- pred domino(domino_matrix::in, irany_matrix::out).
domino(Mx, IMx) :-
	eros_opciok(Opciok),
	domino(Mx, IMx, Opciok).

% Az Mx domin-mtrix ltal adott domin feladat megoldsa az IMx
% irnymtrix. A megoldsi mdszer az Opciok opcilistnak felel meg.
% :- pred domino(domino_matrix::in, irany_matrix::out, varians::in).
domino(Mx, IMx, Varians) :-
	opcio_kifejtes(Varians, Opciok),
	opcio(csakkor, Valt, Opciok),
	irany_korlatok(Mx, IKMx, IKL, Valt),
	domino_korlatok(Mx, IKMx, DVL, Valt),
	cimkezendo_valtozok_opciok(DVL, IKL, Cimkezendo, LOpciok, Opciok),
	labeling(LOpciok, Cimkezendo),
	irany_matrix_konv(IKMx, IMx).

% A DVL domin-vltoz-lista s IKL0 irnykd-vltoz-lista, valamint az
% Opciok opci-lista alapjn meghatrozza a Cimkezendo vltoz-listt
% valamint az LOpciok cmkzsi-opci-listt.
% :- pred cimkezendo_valtozok_opciok(list(domino_lerakasi_sorszam)::in,
%				     list(iranykod)::in, list(int)::out,
%				     labeling_opciok::out, opciok::in).
cimkezendo_valtozok_opciok(DVL, IKL0, Cimkezendo, LOpciok, Opciok) :-
	opcio(szurtek, Ek, Opciok), opcio(szur, Szur, Opciok),
	opcio(label, LOpciok0, Opciok), opcio(valt, Mit, Opciok),
	(   Szur = ki -> LOpciok = LOpciok0, IKL = IKL0
	;   Szur = elott -> filter_vars(IKL0, Ek, IKL), LOpciok = LOpciok0
	;   integer(Szur),
	    filter_vars(IKL0, Ek, IKL), 
	    execution_state(private(Priv)), create_mutable(IKL/1, Priv),
	    LOpciok = [value(szur_helyettesit(Ek,Szur,LOpciok0))|LOpciok0]
	),
	cimkezendo_valtozok(Mit, DVL, IKL, Cimkezendo).

% cimkezendo_valtozok(Mit, DVL, IKL, Cimkezendo): A Mit opcinak
% megfelelen Cimkezendo azonos a DVL domin-vltoz-listval vagy az IKL
% irnykd-vltoz-listval.
% :- pred cimkezendo_valtozok(mit_cimkez::in,
%			      list(domino_lerakasi_sorszam)::in,
%			      list(iranykod)::in, list(int)::out).
cimkezendo_valtozok(domino, DVL, _, DVL).
cimkezendo_valtozok(irany, _, IKL, IKL).

% Szrst vgez az execution_state-ben elmentett irany-vltoz-listn az Ek
% rtkeket szrve minden N-edik meghvs esetn. Ezutn mindenkppen
% behelyettesti a V vltozt az LOpciok cmkzsi opcik szerint.
% :- pred szur_helyettesit(list(int)::in, int::in, labeling_opciok::in,
%			   int, list(int)::in, any::in).
szur_helyettesit(_Ek, _N, _LOpciok, V, _Vk, _BB) :- integer(V), !.
szur_helyettesit(Ek, N, LOpciok, V, _Vk, _BB) :-
	execution_state(private(Priv)), get_mutable(IKL0/I0, Priv),
	(   I0 < N -> I is I0+1, update_mutable(IKL0/I, Priv)
	;   filter_vars(IKL0, Ek, IKL), update_mutable(IKL/1, Priv)
	),
	labeling(LOpciok, [V]).


% Irny-korltok
% --------------

% Az IKMx irnykd-mtrix az Mx domin-mtrixszal azonos mret, elemeinek
% listja az IKL lista, s a szomszdos elemek kztt fennllnak az
% irny-korltok. A korltoknak a Valt ltal elrt vltozatt hasznljuk.
% :- pred irany_korlatok(domino_matrix::in, iranykod_matrix::out,
%                        list(iranykod)::out, csvalt::in).
irany_korlatok(Mx, IKMx, IKL, Valt) :-
	same_size_matrix(Mx, IKMx),
	append_lists(IKMx, IKL),
	domain(IKL, 1, 4),
	irany_matrix_korlatok(IKMx, sor, Valt),
	transpose(IKMx, TIKMx),
	irany_matrix_korlatok(TIKMx, oszlop, Valt).

% irany_matrix_korlatok(IKMx, Tengely, Valt): Az IKMx irnykd-mtrix
% soraiban a szomszdos elemek kztt fennllnak a Tengely (sor;oszlop)
% irnytsnak megfelel irny-korltok. A korltoknak a Valt ltal elrt
% vltozatt hasznljuk. 
% :- type tengely ---> sor ; oszlop.
% :- pred irany_matrix_korlatok(iranykod_matrix::in, tengely::in, csvalt::in).
irany_matrix_korlatok([], _, _).
irany_matrix_korlatok([Sor|Sorok], Tengely, Valt) :-
	irany_lista_korlatok(Sor, 0, Tengely, Valt),
	irany_matrix_korlatok(Sorok, Tengely, Valt).

% irany_lista_korlatok(IKL, EIK, Tengely, Valt): Az [EIK|IKL]
% irnykd-listban a szomszdos elemek kztt fennllnak a Tengely
% irnytsnak megfelel irny-korltok. A korltoknak a Valt ltal elrt
% vltozatt hasznljuk. (EIK = elz irnykd.)
% :- pred irany_lista_korlatok(list(iranykod)::in, iranykod::in,
%                            tengely::in, csvalt::in).
irany_lista_korlatok([], Utso, Tengely, Valt) :-
	szomszed(Tengely, Utso, 0, Valt).
irany_lista_korlatok([E|Ek], Elozo, Tengely, Valt) :-
	szomszed(Tengely, Elozo, E, Valt),
	irany_lista_korlatok(Ek, E, Tengely, Valt).


% Domin-korltok
% ---------------

% Fellltja az Mx domin-mtrixnak s az IKMx irnykd-mtrixnak megfelel
% domin-korltokat, a Valt vltozatban.
% :- pred domino_korlatok(domino_matrix::in, iranykod_matrix::in,
%                         list(domino_lerakasi_sorszam)::out, csvalt::in).
domino_korlatok(Mx, IKMx, DVL, Valt) :-
	bagof(PozIL, Dom^bagof(Poz-IK, lerakas(Mx, Dom, Poz, IK), PozIL), DLk),
	domino_lerakas_korlatok(DLk, IKMx, DVL, Valt).

% domino_lerakas_korlatok(DLk, IKMx, Valt): A DLk domin-leraksoknak s az
% IKMx irnykd-mtrixnak megfelel domin-korltokat fellltja, a Valt
% vltozatban.
% :- pred domino_lerakas_korlatok(lerakasok::in, iranykod_matrix::in,
%                                 list(domino_lerakasi_sorszam)::out,
%				  csvalt::in).
domino_lerakas_korlatok([], _, [], _).
domino_lerakas_korlatok([PozIL|DLk], IKMx, [DV|DVL], Valt) :-
	length(PozIL, Len),
	DV in 1..Len,
	domino_lerakas_korlat(PozIL, DV, 1, IKMx, Valt),
	domino_lerakas_korlatok(DLk, IKMx, DVL, Valt).

% domino_lerakas_korlat(PozIL, DV, N, IKMx, Valt): A DV domin-vltozra
% vonatkoz, a PozIL pozci-irny-lista ltal lert domin-korltokat 
% fellltja, a Valt vltozatban. N a PozIL els
% elemnek sorszma, IKMx az irnykd-mtrix.
% :- pred domino_lerakas_korlat(list(pozirany)::in,
%                domino_lerakasi_sorszam, domino_lerakasi_sorszam::in,
%                iranykod_matrix::in, csvalt::in).
domino_lerakas_korlat([], _, _, _, _).
domino_lerakas_korlat([Poz-IK|PozIL], DV, N, IKMx, Valt) :-
	matrix_elem(Poz, IKMx, IKV),
	csakkor_egyenlo(Valt, DV, N, IKV, IK),
	N1 is N+1,
	domino_lerakas_korlat(PozIL, DV, N1, IKMx, Valt).


% Az irnyok kdolsa
% -------------------

% irany_konv(Sz, I): Az Sz irnykdnak megfelel irny az I.
% :- pred irany_konv(iranykod::out, irany::out).
irany_konv(1, w).
irany_konv(2, n).
irany_konv(3, e).
irany_konv(4, s).

% szomszed(T, E, M, Valt): A T tengely (sor;oszlop) irnyban egyms mellett
% llhat az E els (bal;fels) s M msodik (jobb;als) irnykd.
% A korltot a Valt ltal elrt vltozatban lltja fel.
% :- pred szomszed(tengely::in, iranykod, iranykod, csvalt::in).
szomszed(sor, Bal, Jobb, Valt) :-
	csakkor_egyenlo(Valt, Bal, 1, Jobb, 3).
szomszed(oszlop, Felso, Also, Valt) :-
	csakkor_egyenlo(Valt, Felso, 2, Also, 4).


% Domin leraksok
% ----------------

% Az Mx domin-tblban a Dom domin els mezje lerakhat a Poz 
% pozicira I irnytssal.
% :- pred(domino_matrix::in, domino::out, pozicio::out, iranykod::out)
lerakas(Mx, Dom, Poz, I) :-
	matrix_elem(Poz, Mx, Pottyok1),
	iranyba(I, Poz, Poz1),
	matrix_elem(Poz1, Mx, Pottyok2),
	(   Pottyok1 > Pottyok2 -> Dom = Pottyok2:Pottyok1
	;   Dom = Pottyok1:Pottyok2
	).

% iranyba(I, Poz1, Poz2): Poz1 s Poz2 egy domin kt felnek pozcii,
% s a Poz1-re kerl irnykd I, ahol I=1 vagy I=2.
% :- pred iranyba(iranykod::in, pozicio::in, pozicio::out).
iranyba(1, S*O, S*O1) :- O1 is O+1.
iranyba(2, S*O, S1*O) :- S1 is S+1.

% Az Mx mtrix S. sornak O. eleme E.
% :- pred matrix_elem(pozicio::in, matrix(T)::in, T::out).
matrix_elem(S*O, Mx, E) :-
	nth(S, Mx, Sor),
	nth(O, Sor, E).

% Az irnymtrix konverzija
% --------------------------

% Az IKMx irnykd-mtrixnak megfelel az IMx irnymtrix.
% :- pred irany_matrix_konv(iranykod_matrix::in, irany_matrix::out).
irany_matrix_konv(IKMx, IMx) :-
	findall(IS, (member(IKS, IKMx),
		     findall(I, (member(IK, IKS), irany_konv(IK, I)),
			     IS)),
		IMx).

% A csakkor egyenl korlt
% ------------------------

% csakkor_egyenlo(Valt, X, C, Y, D): Fennll az `X=C akkor s csak
% akkor ha Y=D' korlt, a Valt ltal adott megvalsts szerint.
% :- pred csakkor_egyenlo(csvalt::in, int, int::in, int, int::in).  
csakkor_egyenlo(reif, X, C, Y, D) :-
	X #= C #<=> Y #= D.
csakkor_egyenlo(ind1, X, C, Y, D) :-
        'x=c=>y=d'(X, C, Y, D),
        'x=c=>y=d'(Y, D, X, C).
csakkor_egyenlo(ind2, X, C, Y, D) :-
	'x=c<=>y=d'(X, C, Y, D).

% Fennll az `X=C akkor s csak akkor ha Y=D' korlt.
% :- pred 'x=c<=>y=d'(int, int::in, int, int::in).	
'x=c<=>y=d'(X, C, Y, D) +:
	X in ((dom(Y) /\ {D}) ? (inf..sup) \/ \({C})) /\
	     ((dom(Y) /\ \({D})) ? (inf..sup) \/ {C}),
	Y in ((dom(X) /\ {C}) ? (inf..sup) \/ \({D})) /\
	     ((dom(X) /\ \({C})) ? (inf..sup) \/ {D}).

% Fennll a `ha X=C akkor Y=D' korlt.
% :- pred 'x=c=>y=d'(int, int::in, int, int::in).	
'x=c=>y=d'(X, C, Y, D) +:
	X in (dom(Y) /\ {D}) ? (inf..sup) \/ \({C}),
	Y in ({X} /\  \({C})) ? (inf..sup) \/ {D}.


% Opcikezels
% ------------

% A Nev opci rtke az Opciok opcilista szerint Ertek.
% :- pred opcio(atom::in, any::out, opciok::in).
opcio(Nev, Ertek, Opciok) :-
	member(Nev=Ertek0, Opciok), !, Ertek = Ertek0.
opcio(Nev, Ertek, _) :-
	opcio_alap(Nev, Ertek).

% A Varians egy varinsnv vagy opci-lista, a neki megfelel
% opci-lista Opciok. 
% :- pred opcio_kifejtes(varians::in, opciok::out).
opcio_kifejtes(Varians, Opciok) :-
	atom(Varians), Varians \== [], !, opcio_varians(Varians, _, Opciok).
opcio_kifejtes(Opciok, Opciok).



