% domino.pl llomny
% A domin feladvny 0-1 rk vltozkon alapul megoldsa
% 1.3 Verzi, 2000. aug. 8.
% Nagyhatkonysg Logikai Programozs
% (c) Szeredi Pter

:- module(domino_szp2, [domino/2, domino/3, option_variant/2]).

:- use_module(library(lists)).
:- use_module(library(clpfd)).
:- use_module(library(assoc)).
:- use_module('../utils', [between/3,max_matrix/2]).
:- use_module('../testenv', [filter_vars/3]).

% A megvalsts alapelvei
% ------------------------

% Korlt vltozk:

%    A domin-tblban minden egysgnyi vlasztvonalnak megfeleltetnk egy
%    0-1 rtk korlt-vltozt. Az 1 rtk jelentse az hogy az adott
%    vonalra illeszkedik egy domin kzpvonala. Ha a vltoz 0 rtk,
%    akkor az adott vonal egy domin hatrvonala. rtelemszeren a
%    domin-mtrix szleinek 0 az rtke.

% Korltok:

% Szomszdsgi korltok:
%    A domin-mtrixban minden egysgngyzet ngy oldala kzl pontosan
%    egyet kell lefedni egy domin kzpvonalval. Teht az egysgngyzet
%    oldalainak megfelel korlt-vltozk sszege 1.

% Leraksi korltok:
%    Minden dominhoz tekintjk annak a leraksi lehetsgeit, minden
%    leraks egy kzpvonalat fed le. Miutn mindegyik domin pontosan
%    egyszer kell elforduljon, a leraksok kzpvonalaihoz tartoz
%    korlt-vltozk sszege 1.

% Plda:

% Az albbi feladatmtrixhoz:

% 1 2 1 2
% 0 1 0 2
% 0 1 2 0

% az [A,B,C,...,P,Q] vltozk tartoznak:

%      ----- ----- ----- ----- 
%     |     |     |     |     |
%     |  1  A  2  C  1  E  2  |
%     |     |     |     |     |
%      --B-- --D-- --F-- --G-- 
%     |     |     |     |     |
%     |  0  H  1  J  0  L  2  |
%     |     |     |     |     |
%      --I-- --K-- --M-- --N-- 
%     |     |     |     |     |
%     |  0  O  1  P  2  Q  0  |
%     |     |     |     |     |
%      ----- ----- ----- -----

% Szomszdsgi korltok (az albbi listk mindegyiknek sszege 1):
%                        
% [A,B], [A,D,C], [C,F,E], [E,G],         
% [B,I,H], [D,H,K,J], [F,J,M,L], [G,L,N]  
% [I.O], [K,O,P], [M,P,Q], [N,Q]                               

% Leraksi korltok (az albbi listk mindegyiknek sszege 1):
%
% /*0:0*/ [I]
% /*0:1*/ [B,F,H,J,O]
% /*0:2*/ [L,M,N,Q]
% /*1:1*/ [K]
% /*1:2*/ [A,C,D,E,P]
% /*2:2*/ [G]

% Megjegyzs: az brn '*'-gal jellt helyeken 0 rtk "lvltoz" van.
% Ezeket a nem tartalmazzk a korltoknl mutatott listk.

% Cmkzs:

% A korlt-vltozkat egyetlen listba gyjtjk, felvltva a vzszintes s
% fggleges vonalakat, sorfolytonosan. A sorrend megfelel a fenti brn a
% vltozk ABC-sorrendjnek. Ha a cmkzs vltoz-kivlaszt opcijaknt
% az alaprtelmezst (`leftmost') vlasztjuk, akkor az egymshoz kzeli
% vltozk behelyettestse is egymshoz idben kzel trtnik.  gy egy
% esetleges rossz leraks, ami a krnyezetben egy le nem fedhet mezt
% llt el, viszonylag hamar meghisulshoz vezet.

% Bonyolultabb cmkzsi algoritmusokat valszinleg nem rdemes hasznlni.
% A first-fail elv 0-1-rtk vltozkat nem klnbzteti meg (minden
% nem-behelyettestett vltoz tartomnya azonos mret, ktelem). Az
% 'ffc' cmkzsi md viszont kifejezetten lassthat, mert egy rossz
% dntst kveten a tbla ms rszein sok-sok vlasztsi pontot hozhat
% ltre, mieltt visszatr a rossz dnts krnyezetbe.

% A cmkzs sorn fontos szerepet kap az n. szrs. Szrskor adott egy
% vltoz-lista s rtkek egy listja. Sorra vesszk a vltozkat s
% mindegyiket megprbljuk mindegyik rtkre behelyettesteni. Ha egy
% behelyettests meghisulst okoz (kzvetlenl, tovbbi cmkzs nlkl),
% akkor az adott rtket kivesszk a vltoz tartomnybl, egybknt nem
% csinlunk semmit. Fontos megjegyezni, hogy a szrs nem hoz ltre
% vlasztsi pontot.

% Ebben a megoldsban rtelemszeren a hatr-vltozkat szrjk, mgpedig
% alaprtelmezsben csak az [1] rtk-listra. Igaz ugyanis, hogy ha egy H1
% hatr-vltoz 0-ra helyettestse meghiuslst okoz, akkor a vele egy
% egysgngyzet alkot H2, H3, H4 hatr-vltozk 1-re helyettestse is
% meghisulst fog okozni. Teht az 1-esek szrsekor H2, H3 s H4
% mindegyike 0-val helyettestdik, aminek kvetkeztben, a megfelel
% szomszdsgi korlt alapjn H1 1 rtket kap. Pontosan ugyanezt a hatst
% rn el a H1-nek 0 rtkkel val szrse. Ezzel megmutattuk, hogy a
% [0,1] rtk-listval val szrs nem tud jobban szkteni mint az [1]
% listval val szrs.

% A cmkzs s szrs lefolyst opcikkal szablyozhatjuk: pl.
% megadhatjuk a labeling/2 cmkzsi opciit.  Alaprtelmezsben nem
% szrnk, de opcival elrhatjuk, hogy szrjnk, s azt is, hogy milyen
% gyakorisggal (meghatrozhatjuk, hogy a labeling sorn csak minden
% N. vltoz esetn szrjnk). Megvltoztathatjuk a szrskor hasznland
% rtk-listt is.

% A program vzlata

% Fellltjuk a leraks- majd a szomszdsgi korltokat. A leraks-korltok
% fellltshoz elszr minden dominhoz kigyjtjk a lehetsges leraksok
% kzpvonalhoz tartoz hatr-vltozkat. A kigyjtshez a library(assoc)
% knyvtr eljrsait hasznljuk. Vgigmegynk a domin-tbln, ezzel
% prhuzamosan ptve a hatr-mtrixot. Minden mezre megvizsgljuk a r
% vzszintesen jobbra ill. a fgglegesen lefel lerakhat domint, s a
% domin kzpvonalhoz rendelt hatr-vltozt a dominhoz asszociljuk. A
% kigyjts befejezse utn vgigmegynk az asszocicis fn s minden
% dominhoz a kigyjttt hatrvltoz-listra felvesszk a megfelel
% leraks-korltot.

% A szomszdsgi korltok fellltshoz a hatr-mtrixon megynk vgig s
% minden egysgngyzetre felvesszk a szomszdsgi korltot.

% Vgl a hatr-mtrixbeli vltozlistn elvgezzk a cmkzst (belertve
% a szrst), majd a kapott megoldsra a hatrvltozk alapjn felptjk
% az irny-mtrixot.

% Adattpusok
% -----------

% :- type hatar_matrix == matrix(hatarpar).
% :- type hatarpar ---> kelet/del.   % a mtrix minden mezjhez a keleti
%                                    % s dli hatrvonalat troljuk
% :- type kelet == hatar.
% :- type del == hatar.
% :- type hatar == int.                      % 0 vagy 1

% :- type domino_matrix == matrix(pottyok).
% :- type domino ---> pottyok:pottyok
% :- type pottyok == int.

% :- type irany_matrix == matrix(irany).
% :- type irany ---> w ; n ; e ; s.

% :- type opciok == list(any).

% :- type matrix(Type) == list(list(Type)).

% :- type assoc(Key,Val) ---> % A library(assoc) ltal definilt
%			      % AVL asszocicis fa tpusa
%			      t ;
%                             t(Key,Val,int,assoc(Key,Val),assoc(Key,Val)).

% :- type asszoc_fa == assoc(domino, list(hatar)).
%                      % A leraksi korltok kigyjtshez hasznlt fa.


% Algoritmus-vltozatok
% ---------------------

% A domino/3 belpsi pont harmadik, opcilista argumentuma segtsgvel az
% algoritmus klnfle vltozatait prblhatjuk ki;

% Az opcilista lehetsges elemei:

% osszeg=sum (alaprt.)		a lista_osszege_1 felttelt a sum/3 korlt
%				segtsgvel valstjuk meg.
% osszeg=ari(N)			A lista_osszege_1 felttelt N-nl nem hosszabb
%				listkra aritmetikai korlt segtsgvel 
%				valstjuk meg, egybknt pedig a sum/3
%				korlttal (N =< 8).
% osszeg=ind(N)			A lista_osszege_1 felttelt N-nl nem hosszabb
%				listkra indexiklissal valstjuk meg,
%				egybknt pedig a sum/3 korlttal (N =< 8).

% :- type hogyan ---> sum ; ari(int); ind(int).

% A fenti vlasztsok az sszes korltra vonatkoznak. Lehetsg van a
% ktfle (szomszdsgi s leraksi) korlt megvalstsnak kln-kln
% val szablyozsra:

% szomsz=Hogyan			A szomszdsgi korltokban a lista_osszege_1
%				felttelt az osszeg=Hogyan esetre lert
%				mdon valstjuk meg.
% lerak=Hogyan			A leraksi korltokban a lista_osszege_1
%				felttelt az osszeg=Hogyan esetre lert
%				mdon valstjuk meg.

% :- type hol ---> szomsz ; lerak.

% szur=ki (alaprt.)           	Nincs szrs.
% szur=elott                	Csak a cmkzs eltt szrnk.
% szur=N	        	A cmkzs eltt szrnk, majd minden N. 
% 				vltoz cmkzse utn ismt.

% szurtek=Ek (alaprt. [1])	Szrskor az Ek egsz-lista elemeit prbljuk
% 				behelyettesteni.

% label=LOpciok			LOpciok megadja a labeling/2 eljrsnak 
%  (alaprt. [])		tovbbadand opcikat.

% :- type labeling_opciok == list(any).

% Alaprtelmezsek

% Az opcik alaprtelmezst gy vlasztottuk meg, hogy azok egy viszonylag
% egyszer megoldsnak feleljenek meg. A domino(Mx, IMx, []) hvs teht
% ezt az egyszer megoldst futtatja. A domino(Mx, IMx) hvst viszont nem
% az elzre vezetjk vissza, hanem egy olyan opci-kszletet hasznlunk,
% amely a nehz feladatok megoldsnl hasznos.

% opcio_alap(Nev, Ertek): A Nev nev opci alaprtelmezse Ertek.
% :- pred opcio_alap(atom::in, any::out).
% opcio_alap(Nev, Ertek): A Nev opci alaprtelmezse Ertek.
% :- pred opcio_alap(atom::in, any::out).
opcio_alap(osszeg,  sum).
opcio_alap(szur,    ki).
opcio_alap(szurtek, [1]).
opcio_alap(label,   []).

% eros_opciok(Opciok): Opciok a domino/2-ben hasznland opcilista.
% :- pred eros_opciok(opciok::out).
eros_opciok([osszeg=ind(5),szur=2]).

% Nhny opcilista-varinst kivlasztottunk s nvvel lttunk el. Ennek
% szerepe ketts: a varinsnv megadhat a domino/3 harmadik
% argumentumaknt. Ezen fell a keretprogram, ha nem egy konkrt varins
% futtatsra krik fel, az sszes megnevezett varinst futtatja (ezeket az
% option_variant/2 exportlt prediktum sorolja fel).

% :- type varians == any.  % opcilista vagy atom
% :- type szint ---> easy ; diff ; hard.

% Varians egy opci-varins neve.
% :- pred option_variant(atom::out, szint::out).
option_variant(Variant, Level) :-
	opcio_varians(Variant, Level, _).


% opcio_varians(Varians, Szint, Opciok): A Varians varinshoz tartoz 
% opci-lista Opciok. Ezt a varinst legfeljebb a Szint szint feladatokra
% rdemes futtatni.
% :- pred opcio_varians(atom::out, szint::out, opciok::out).

opcio_varians(ind5_1,      diff, [osszeg=ind(5),szur=1]).
opcio_varians(ind5_2,      hard, [osszeg=ind(5),szur=2]).
opcio_varians(ind5_3,      hard, [osszeg=ind(5),szur=3]).
opcio_varians(ind5_4,      hard, [osszeg=ind(5),szur=4]).
opcio_varians(ind5_5,      hard, [osszeg=ind(5),szur=5]).

opcio_varians(sum_2,       diff, [osszeg=sum,szur=2]).
opcio_varians(ari5_2,      diff, [osszeg=ari(5),szur=2]).
opcio_varians(ind5_2_sz0,  diff, [osszeg=ind(5),szur=2,szurtek=[0]]).
opcio_varians(ind5_2_sz01, diff, [osszeg=ind(5),szur=2,szurtek=[0,1]]).
opcio_varians(ind5_2_ff,   diff, [osszeg=ind(5),szur=2,label=[ff]]).
opcio_varians(ind5_2_ffc,  diff, [osszeg=ind(5),szur=2,label=[ffc]]).

opcio_varians(ind2_2,      diff, [osszeg=ind(2),szur=2]).
opcio_varians(ind3_2,      diff, [osszeg=ind(3),szur=2]).
opcio_varians(ind4_2,      diff, [osszeg=ind(4),szur=2]).
opcio_varians(ind6_2,      diff, [osszeg=ind(6),szur=2]).

opcio_varians(ind5_elott,  easy, [osszeg=ind(5),szur=elott]).
opcio_varians(ind5_ki,     easy, [osszeg=ind(5),szur=ki]).

opcio_varians(l_ind_sz_sum_ki,easy, [lerak=ind(5),szomsz=sum,szur=ki]).
opcio_varians(l_sum_sz_ind_ki,easy, [lerak=sum,szomsz=ind(5),szur=ki]).
opcio_varians(sum_ki,         easy, [osszeg=sum,             szur=ki]).



% Fbejratok
% -----------

:- public domino/2, domino/3.

% Az Mx domin-mtrix ltal adott domin feladat megoldsa az IMx
% irnymtrix.
% :- pred domino(domino_matrix::in, irany_matrix::out).
domino(Mx, IMx) :-
	eros_opciok(Opciok),
	domino(Mx, IMx, Opciok).

% Az Mx domin-mtrix ltal adott domin feladat megoldsa az IMx
% irnymtrix. A megoldsi mdszer az Opciok opcilistnak felel meg.
% :- pred domino(domino_matrix::in, irany_matrix::out, opciok::in).
domino(Mx, IMx, Opciok0) :-
	opcio_kifejtes(Opciok0, Opciok),
	lerakasi_korlatok(Mx, HMx, Opciok),
	szomszedsagi_korlatok(HMx, HL0, Opciok),
	cimkezendo_valtozok_opciok(HL0, HL, LOpciok, Opciok),
	labeling(LOpciok, HL),
	hatar_matrix_konv(HMx, IMx).

% A DVL hatr-vltoz-lista s az Opciok opci-lista alapjn meghatrozza a
% HL cmkzend vltoz-listt valamint az LOpciok cmkzsi-opci-listt.
% :- pred cimkezendo_valtozok_opciok(list(hatar)::in, list(hatar)::out,
%				     labeling_opciok::out, opciok::in).
cimkezendo_valtozok_opciok(HL0, HL, LOpciok, Opciok) :-
	opcio(szurtek, Ek, Opciok), opcio(szur, Szur, Opciok),
	opcio(label, LOpciok0, Opciok),
	(   Szur = ki -> LOpciok = LOpciok0, HL = HL0
	;   Szur = elott -> filter_vars(HL0, Ek, HL), LOpciok = LOpciok0
	;   integer(Szur), filter_vars(HL0, Ek, HL), 
	    execution_state(private(Priv)), create_mutable(1, Priv),
	    LOpciok = [value(szur_helyettesit(Ek,Szur,LOpciok0))|LOpciok0]
	).

% Szrst vgez a [H|HL] hatr-vltoz-listn az Ek rtkeket szrve minden
% N-edik meghvs esetn. Ezutn mindenkppen behelyettesti a H vltozt
% az LOpciok cmkzsi opcik szerint.
% :- pred szur_helyettesit(list(int)::in, int::in, labeling_opciok::in,
%			   hatar::in, list(hatar)::in, any::in).
szur_helyettesit(_Ek, _N, _LOpciok, H, _HL, _BB) :- integer(H), !.
szur_helyettesit(Ek, N, LOpciok, H, HL, _BB) :-
	execution_state(private(Priv)), get_mutable(I0, Priv),
	(   I0 < N -> I is I0+1, update_mutable(I, Priv)
	;   filter_vars([H|HL], Ek, _), update_mutable(1, Priv)
	),
	labeling(LOpciok, [H]).


% Domin-korltok
% ---------------

% A HMx hatr-mtrix az Mx domin-mtrixszal azonos mret. HMx elemei
% kztt fennllnak a domin-korltok, az Opciok opcilistnak megfelel mdon.
% :- pred lerakasi_korlatok(domino_matrix::in, hatar_matrix::out, opciok::in).
lerakasi_korlatok(Mx, HMx, Opciok) :-
	max_matrix(Mx, Max),
	findall((P1:P2)-[], (between(0, Max, P1), between(P1, Max, P2)), Ures),
	ord_list_to_assoc(Ures, DLk0),
	Mx = [Sor|Sorok],
	domino_lerakasok(Sorok, Sor, HMx, DLk0, DLk),
	melyik_osszeg_korlat(Opciok, lerak, Hogyan),
	fabeli_lerakasi_korlatok(DLk, Hogyan).

% fabeli_lerakasi_korlatok(AsszocFa, Hogyan): Az AsszocFa fa leveleiben
% lev listkra fennllnak a domin-korltok. A korltok fellltst a
% Hogyan opci szerint vgezzk.
% :- pred fabeli_lerakasi_korlatok(asszoc_fa::in, hogyan::in).
fabeli_lerakasi_korlatok(t, _Hogyan).
fabeli_lerakasi_korlatok(t(_,Vk,_,Bal,Jobb), Hogyan) :-
	fabeli_lerakasi_korlatok(Bal, Hogyan),
	length(Vk, Hossz),
	lista_osszege_1(Hogyan, Hossz, Vk),
	fabeli_lerakasi_korlatok(Jobb, Hogyan).

% domino_lerakasok(Mx, Sor, HMx, DLk0, DLk): A HMx hatr-mtrix a [Sor|Mx]
% domin-mtrixszal azonos mret. Ha a DLk0 asszocicis fhoz 
% hozzvesszk a [Sor|Mx] mtrixban lev domin -> kzpvonal-vltoz
% hozzrendelseket, akkor a DLk ft kapjuk.
% :- pred domino_lerakasok(domino_matrix::in, list(pottyok)::in,
%                          hatar_matrix::out aszoc_fa::in, asszoc_fa::out).
domino_lerakasok([KovSor|Sorok], Sor, [HSor|HMx], DLk0, DLk) :-
	domino_sor_lerakasok(Sor, KovSor, HSor, DLk0, DLk1),
	domino_lerakasok(Sorok, KovSor, HMx, DLk1, DLk).
domino_lerakasok([], Sor, [HSor], DLk0, DLk) :-
	KovSor = [-1|KovSor],
	domino_sor_lerakasok(Sor, KovSor, HSor, DLk0, DLk).

% domino_lerakasok(Sor, KovSor, HSor, DLk0, DLk): A HSor hatrpr-sor a Sor
% domin-mtrixsorral azonos mret. Ha a DLk0 asszocicis fhoz
% hozzvesszk a [Sor,KovSor] rszmtrixban lev domin -> kzpvonal-vltoz
% hozzrendelseket, akkor a DLk ft kapjuk.
% :- pred domino_sor_lerakasok(list(pottyok)::in, list(pottyok)::in,
%                        list(hatarpar)::out, asszoc_fa::in, asszoc_fa::out).
domino_sor_lerakasok([Elem|Sor], [Del|KovSor], [HKelet/HDel|HSor], DLk0, DLk) :-
	domino_hozza(Del, Elem, HDel, DLk0, DLk1),
	(   Sor = [Kelet|_] ->
	    domino_hozza(Elem, Kelet, HKelet, DLk1, DLk2),
	    domino_sor_lerakasok(Sor, KovSor, HSor, DLk2, DLk)
	;   DLk = DLk1, HSor = [], HKelet = 0
	).

% domino_hozza(Pottyok1, Pottyok2, Hatar, DLk0, DLk): Ha a DLk0
% asszocicis fhoz hozzvesszk Pottyok1 s Pottyok2 pttykbl ll
% dominnak a Hatar kzpvonal-vltozhoz val hozzrendelst, akkor a DLk
% ft kapjuk.
% :- pred domino_hozza(pottyok::in, pottyok::in, hatar,
%                      asszoc_fa::in, asszoc_fa::out).
domino_hozza(-1, _, Hatar, DLk0, DLk) :- !, DLk = DLk0, Hatar = 0.
domino_hozza(Pottyok1, Pottyok2, Hatar, DLk0, DLk) :-
	Pottyok1 > Pottyok2, !,
	get_assoc(Pottyok2:Pottyok1, DLk0, L, DLk, [Hatar|L]).
domino_hozza(Pottyok1, Pottyok2, Hatar, DLk0, DLk) :-
	get_assoc(Pottyok1:Pottyok2, DLk0, L, DLk, [Hatar|L]).


% Szomszdsgi korltok
% ---------------------

% A HMx hatr-mtrix elemeinek listja a HL lista, s a mtrix minden
% minden mezjnek hatraira fennll a szomszdsgi korlt. A korltok
% fellltst az Opciok opcilista szerint vgezzk.
% :- pred szomszedsagi_korlatok(hatar_matrix::in, list(hatar)::out,
%                               opciok::in).
szomszedsagi_korlatok(HMx, HL, Opciok) :-
	ESor = [0|ESor],
	melyik_osszeg_korlat(Opciok, szomsz, Hogyan),
	hatar_matrix(HMx, ESor, HL, Hogyan).

% hatar_matrix(HMx, ESor, HL, Opciok): A HMx hatr-mtrix elemeinek
% listja a HL lista, s a mtrix minden mezjnek hatraira fennll a
% szomszdsgi korlt. Az ESor listnak kezdszelete a HMx szaki
% hatrvonalainak listja.  (Azrt nem azonos azzal, mert gy ESor kezd-
% ill. vg-rtkeknt a csupa 0 elembl ll vgtelen ciklikus listt
% adhatjuk meg.) A korltok fellltst a Hogyan opci szerint vgezzk.
% :- pred hatar_matrix(hatar_matrix::in,list(hatar)::in, list(hatar)::out,
%                      hogyan::in). 
hatar_matrix([], ESor, [], _Hogyan) :-
	ESor = [0|ESor].
hatar_matrix([HSor|HMx], ESor, HL, Hogyan) :-
	hatar_sor(HSor, 0, ESor, DSor, HL0, HL, Hogyan),
	hatar_matrix(HMx, DSor, HL0, Hogyan).

% hatar_sor(HSorS, Ny, ESor, DSor, HL0, HL, Hogyan): A HSor hatr-sor
% elemeinek listja a HL-HL0 lista, s a sor minden mezjnek hatraira
% fennll a szomszdsgi korlt. Ny a sort megelz mez nyugati hatra. Az
% ESor ill. DSor listknak kezdszelete a HSor szaki ill. dli
% hatrvonalainak listja. A korltok fellltst a Hogyan vltozat szerint
% vgezzk. 
% :- pred hatar_sor(list(hatarpar)::in, hatar::in, 
%                   list(hatar)::in, list(hatar)::out,
%		    list(hatar), list(hatar), hogyan::in).
hatar_sor([], 0, _ESor, _DSor, HL, HL, _Hogyan).
hatar_sor([Kelet/Del|HSor], Nyugat, [Eszak|ESor], [Del|DSor],
	  HL0, [Kelet,Del|HL], Hogyan) :-
	Kelet in 0..1, Del in 0..1,
	lista_osszege_1(Hogyan, 4, [Eszak,Kelet,Del,Nyugat]),
	hatar_sor(HSor, Kelet, ESor, DSor, HL0, HL, Hogyan).


% A hatr-mtrix konverzija
% ---------------------------

% hatar_matrix_konv(HMx, IMx): A HMx hatr-mtrixnak megfelel az IMx
% irny-mtrix. 
% :- pred hatar_matrix_konv(hatar_matrix::in, irany_matrix::out).
hatar_matrix_konv([HSor|HMx], [ISor|IMx]) :-
	hatar_sor_konv(HSor, 0, ISor),
	hatar_matrix_konv(HMx, IMx).
hatar_matrix_konv([], []).

% hatar_sor_konv(HSor, Nyugat, ISor): A HSor hatr-sornak megfelel az ISor 
% irny-sor, ha a HSor-t megelz nyugati hatrvonal Nyugat.
% :- pred hatar_matrix_konv(list(hatarpar)::in, hatar(in), list(irany)::out).
hatar_sor_konv([Kelet/Del|HSor], Nyugat, [I|ISor]) :-
	Kod is Nyugat<<2+Del<<1+Kelet, kod_irany(Kod, I),
	hatar_sor_konv(HSor, Kelet, ISor).
hatar_sor_konv([], _, []).

% kod_irany(Kod, Irany): Kod egy olyan egsz szm, amely egy mez hrom
% oldala hatrvltozjnak rtkt tartalmazza binris szm formjban
% (nyugati, dli s keleti hatrok, rendre a 4, 2, 1 helyirtkeken). Irany
% a Kod kd mez irnya (az 1 rtk hatrral szemkzti gtj).

% :- pred kod_irany(int::in, irany::out).
%           WSE
kod_irany(2'000, s).
kod_irany(2'010, n).
kod_irany(2'001, w).
kod_irany(2'100, e).


% sszeg-korlt vltozatok
% ------------------------

% Hogyan a Hol fajtj korlt megvalstst elr opci, az Opciok
% opcilistnak megfelelen.
% :- pred melyik_osszeg_korlat(opciok::in, hol::in, hogyan::out).
melyik_osszeg_korlat(Opciok, Hol, Hogyan) :-
	member(Hol=Hogyan0, Opciok), !, Hogyan = Hogyan0.
melyik_osszeg_korlat(Opciok, _, Hogyan) :-
	opcio(osszeg, Hogyan, Opciok).

% lista_osszege_1(Hogyan, Hossz, Lista): Lista egy Hossz hossz
% korltvltoz lista, amelynek sszege 1. A korltot a Hogyan opcinak
% megfelel mdon lltja fel.
lista_osszege_1(ari(N), H, Lista) :- 
	H =< N, !, osszeg1_ari(H, Lista).
lista_osszege_1(ind(N), H, Lista) :-
	H =< N, !, osszeg1_ind(H, Lista).
lista_osszege_1(_, _H, Lista) :-
	sum(Lista, #=, 1).

% osszeg1_ind(Hossz, L): L egy Hossz hossz lista, amelynek sszege 1. A
% korltot FD-prediktum segtsgvel lltja fel ha Hossz =< 8, klnben
% a sum/3 korlttal.
% :- pred osszeg1_ind(int::in, list(hatar)::in).
osszeg1_ind(1, L) :- !, L = [1].
osszeg1_ind(2, L) :- !, L = [A,B],             o1fd(A, B).
osszeg1_ind(3, L) :- !, L = [A,B,C],           o1fd(A, B, C).
osszeg1_ind(4, L) :- !, L = [A,B,C,D],         o1fd(A, B, C, D).
osszeg1_ind(5, L) :- !, L = [A,B,C,D,E],       o1fd(A, B, C, D, E).
osszeg1_ind(6, L) :- !, L = [A,B,C,D,E,F],     o1fd(A, B, C, D, E, F).
osszeg1_ind(7, L) :- !, L = [A,B,C,D,E,F,G],   o1fd(A, B, C, D, E, F, G).
osszeg1_ind(8, L) :- !, L = [A,B,C,D,E,F,G,H], o1fd(A, B, C, D, E, F, G, H).
osszeg1_ind(_, L) :- sum(L, #=, 1).

% osszeg1_ari(Hossz, L): L egy Hossz hossz lista, amelynek sszege 1. A
% felttelt aritmetikai korlt segtsgvel lltja fel ha Hossz =< 8,
% klnben a sum/3 korlttal.
% :- pred osszeg1_ari(int::in, list(hatar)::in).
osszeg1_ari(1, L) :- !, L = [1].
osszeg1_ari(2, L) :- !, L = [A,B],             A+B #= 1.
osszeg1_ari(3, L) :- !, L = [A,B,C],           A+B+C #= 1.
osszeg1_ari(4, L) :- !, L = [A,B,C,D],         A+B+C+D #= 1.
osszeg1_ari(5, L) :- !, L = [A,B,C,D,E],       A+B+C+D+E #= 1.
osszeg1_ari(6, L) :- !, L = [A,B,C,D,E,F],     A+B+C+D+E+F #= 1.
osszeg1_ari(7, L) :- !, L = [A,B,C,D,E,F,G],   A+B+C+D+E+F+G #= 1.
osszeg1_ari(8, L) :- !, L = [A,B,C,D,E,F,G,H], A+B+C+D+E+F+G+H #= 1.
osszeg1_ari(_, L) :- sum(L, #=, 1).

% o1fd(A, B, ...): A+B+... = 1 fennll.
o1fd(A, B) +:                                  A+B #= 1.
o1fd(A, B, C) +:                               A+B+C #= 1.
o1fd(A, B, C, D) +:                            A+B+C+D #= 1.
o1fd(A, B, C, D, E) +:                         A+B+C+D+E #= 1.
o1fd(A, B, C, D, E, F) +:                      A+B+C+D+E+F #= 1.
o1fd(A, B, C, D, E, F, G) +:                   A+B+C+D+E+F+G #= 1.
o1fd(A, B, C, D, E, F, G, H) +:                A+B+C+D+E+F+G+H #= 1.


% Opcikezels
% ------------

% A Nev opci rtke az Opciok opcilista szerint Ertek.
% :- pred opcio(atom::in, any::out, opciok::in).
opcio(Nev, Ertek, Opciok) :-
	member(Nev=Ertek0, Opciok), !, Ertek = Ertek0.
opcio(Nev, Ertek, _) :-
	opcio_alap(Nev, Ertek).

% A Varians egy varinsnv vagy opci-lista, a neki megfelel
% opci-lista Opciok. 
% :- pred opcio_kifejtes(varians::in, opciok::out).
opcio_kifejtes(Varians, Opciok) :-
	atom(Varians), Varians \== [], !, opcio_varians(Varians, _, Opciok).
opcio_kifejtes(Opciok, Opciok).
