:- ensure_loaded(time_log).
% A time_log.pl llomny
%
%		results(P/V, Test, Rs)
%
% alak tnylltsokat tartalmaz, ahol P a programvltozat, V annak egy
% opci-varinsa, Test egy teszt-eset, Rs pedig Nev=Ertek alak prok
% listja. Ebbl itt csak az msec=Ido alaku prt hasznljuk, amely az adott
% teszt-eset adott varinson val futsnak ideje. (Az rtkelsben a P
% programvltozat-nevet elhagyjuk, mert a V nv egyrtelmen azonostja P-t
% is.) 

:- use_module(library(lists)).

% A Test-re illeszthet teszt-esetekre vonatkozan osztlyozza a
% varinsokat, s az eredmnyt megjelenti.
% Csak azokat a teszt-eseteket vesszk, figyelembe, amelyek a leggyorsabb
% vltozatban is legalbb a time_limit tnylltsban megadott ideig
% (jelenleg 100 ms) futottak.
classify_variants(Test) :-
	save_times(Test),
	setof(TotalTime-Class/V,
	      Test^classify_variant(Test, V, Class, TotalTime), CVs),
	(   member(TT-(NGoods0-Wins)/V, CVs),
	    format('~N~w~t~16|', [V]),
	    (   TT == partial -> format('partially tested', [])
	    ;   format('total time ~3d sec', [TT])
	    ),
	    nl,
	    (   NGoods0 < 0 -> NGoods is -NGoods0,
		format('~t~16|within 5% of the best on ~d tests~n', [NGoods])
	    ;   true
	    ),
	    reverse(Wins, RWins), member(P-T, RWins),
	    format('~t~16|winner on: ~w with the lead of ~d%~n', [T,P]),
	    fail
	;   true
	).

% A Test-re illeszthet teszt-esetek kztt a V varins a Wins
% teszt-esetekben az els helyen vgzett, ezeket is belertve NGoods
% esetben pedig a legjobbtl legfeljebb 5%-kal trt el. A V varins
% sszestett futsi ideje TotalTime. Ez a 'partial' atommal egyenl, ha
% az adott varins nem az sszes teszt-esetre futott.
classify_variant(Test, V, NGoods-Wins, TotalTime) :-
	count_tests(Test, NumAllTests), 
	all_variants(Vs), member(V, Vs),
	(   setof(Test, good_variant(V, Test), Goods) ->
	    length(Goods, NGoods0)
	;   NGoods0 = 0
	),
	NGoods is -NGoods0,
	(   setof(P-Test, winner(V, Test, P), Wins) -> true
	;   Wins = []
	),
	findall(T, (times(Test,TVs),member(T-V,TVs)), Times), Times \== [],
	(   length(Times, NumAllTests) -> sum_list(Times, TotalTime)
	;   TotalTime = partial
	).

% A Test-re illeszthet sszes teszt-eset szma NumAllTests.
count_tests(Test, NumAllTests) :-
	setof(Test, Ts^times(Test, Ts), AllTests),
	length(AllTests, NumAllTests).


:- dynamic times/2.

% A Test-re illeszthet teszt-esetekre ltrehozza a times0/2 prediktum egy
% dinamikus msolatt times/2 nven.
save_times(Test) :-
	retractall(times(_,_)),
	times0(Test, Times), assertz(times(Test, Times)), fail.
save_times(_).

% A Test teszt-eset futsi-id listja Ts. A Ts lista Id-Vltozatnv
% prokbl ll, nvekven rendezve. A Test teszt-eset legjobb futsi ideje
% is legalbb a time_limit tnylltsban megadott szm ezredmsodperc.
times0(Test, Times) :-
	time_limit(Limit), 
	setof(Time-V, test_time(Test, V, Time), Times),
	Times = [Best-_|_], Best >= Limit.

% A Test teszt-eset futsi ideje a Variant opci-varins esetn Time
% ezredmsodperc. 
test_time(Test, Variant, Time) :-
	results(V, Test, Rs),
	V = _/Variant,
	member(msec=Time, Rs).

% Vs az sszes varinsnv listja.
all_variants(Vs) :-
	findall(V, test_time(_, V, _), Vs0),
	remove_duplicates(Vs0, Vs).

% A vizsgland tesztesetek futsi idejnek als hatra.
time_limit(100).

% A V varinssal fut a Test teszt-eset a leggyorsabban, a msodik
% helyezettnl LeadPercent szzalkkal gyorsabban.
winner(V, Test, LeadPercent) :-
	times(Test, Ts), Ts = [T0-V,T1-_|_],
	LeadPercent is T1*100//T0-100.

% A V varins a Test teszt-esetre a legjobb varinstl legfeljebb 5
% szzalkkal marad el.
good_variant(V, Test) :-
	times(Test, Ts), Ts = [T0-_|_],
	member(T-V, Ts), T =< 1.05*T0.


