:- module(tests, [run_all/0, run_all/1, run_all/2, run_all_max/2, run/2,
		  clear/0, cmp/3, reset_time_log/0,
		  filter_vars/3]).

:- use_module([library(lists),library(system),library(clpfd)]).
:- use_module(library(timeout)).
:- ensure_loaded([kdomino]).

% Konfigurcis dolgok %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Idkorlt ezredmsodpercben.
:- bb_put(user:time_out, 7200000).

% variant(N, V): V a domino feladat egy megoldst tartalmaz
% knyvtr, N pedig ennek a varinsnak a rvid azonostja.
variant(gsala, 'SalamonGabor').
variant(isti, 'BertaIstvan').
variant(manusz, 'MannZoltan').
variant(ujpi, 'MahrTamas').
variant(tpeti, 'TarjanPeter').
variant(mirko, 'VisontaiMirko').
variant(inti, 'OrbanAndras').
variant(karesz, 'OporKaroly').
variant(gergo, 'VarroGergely').
variant(ggabor, 'GesztesiGabor').
variant(dtamas, 'DobosTamas').
variant(csaba, 'GasparCsaba').

variant(szp, 'SzerediPeter').
variant(szp2, 'SzerediPeter2').

% Belpsi pontok %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Az sszes vltozatot az sszes lehetsges opcival futtatja az sszes
% engedlyezett tesztadatra.  Ha a megolds definilja az option_variant/2
% eljrst, akkor az azltal megadott sszes opcival futtatja az
% engedlyezett nehzsgi szintekre.  Ha ha nem definilja, akkor az sszes
% ltez tesztadat engedlyezett.  A szabvnyos kimenetre statisztikt r.
% A time_log fjlhoz hozzrja a futsrl szl napl-bejegyzst.
run_all :-
	run_all(_).

% A V0-ra illeszthet sszes varinst az sszes lehetsges opcival
% futtatja az sszes engedlyezett tesztadatra.  A szabvnyos kimenetre
% statisztikt r.  A time_log fjlhoz hozzrja a futasrl szl
% napl-bejegyzst.
run_all(V0) :-
	test_case(V0, V, Level),
	not_above(Level, TstDir),
	run(V, TstDir/_), fail.
run_all(_).

% A V0-ra illeszthet sszes varinst az sszes lehetsges opcival
% futtatja az sszes olyan engedlyezett tesztadatra, ahol a nehzsgi
% szint nem magasabb mint MaxLevel.  A szabvnyos kimenetre statisztikt
% r.  A time_log fjlhoz hozzrja a futasrl szl napl-bejegyzst.
run_all_max(V0, MaxLevel) :-
	test_case(V0, V, Level),
	not_above(Level, TstDir),
	not_above(MaxLevel, TstDir),
	run(V, TstDir/_), fail.
run_all_max(_, _).

% A V0-ra illeszthet sszes vltozatot az sszes lehetsges opcival
% futtatja a Tst ltal megadott sszes tesztadatra.  A szabvnyos 
% kimenetre statisztikt r.  A time_log fjlhoz hozzrja a futasrl
% szl napl-bejegyzst.
run_all(V0, Tst) :-
	test_case(V0, V, _),
	run(V, Tst), fail.
run_all(_, _).

% A V vltozatot futtatja a Tst ltal megadott sszes tesztadatra.
% A szabvnyos kimenetre statisztikt r.
% A time_log fjlhoz hozzrja a futasrl szl napl-bejegyzst.
run(V, Tst) :-
	load_variant(V, _),
	write('Running variant '), write(V), nl,
	flush_output,
	open(time_log, append, S),
	bb_put(total_counters, [msec=0,sols=0]),
	run2(Tst, V, S),
	bb_get(total_counters, TCnts0),
	select(tests=NoTests, TCnts0, TCnts),
	(   ground(Tst), NoTests =< 1 -> true
	;   write('----------------'), nl,
	    Tst = Group/_,
	    show_results(Group/'TOTAL', V, TCnts), nl,
	    writeq(S, total_results(V,Group,TCnts)), write(S, '.\n'),
	    nl(S)
	),
	close(S).

% A V1 s a V2 vltozatok megoldsait sszehasonltja a T ltal jellt
% teszt-esetekre.  Ha eltrst tall, szl.
cmp(V1, V2, T) :-
	load_variant(V1, _),
	load_variant(V2, _),
	format('Comparing variants ~w and ~w.~n', [V1, V2]),
	pr(T, M),
	format('Test ~w...~t~30|', [T]),
	flush_output,
	fail_if_time_out(domino_sols(V1, M, Sols1), V1),
	fail_if_time_out(domino_sols(V2, M, Sols2), V2),
	(   eq_sols(Sols1, Sols2) -> write(' no differences found.\n')
	;   format('~| ~`*t DIFFERENCES FOUND ~`*t~40+~n', [])
	),
	fail.
cmp(_, _, _).

% 0 mretre csonkolja a time_log fjlt.
reset_time_log :-
	open(time_log, write, S), close(S).


% Segdeljrsok %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% V egy V0-ra illeszthet vltozat, amely Level nehzsgi szint.
test_case(V0, V, Level) :-
	(   variant(V0, V1) -> true
	;   atom(V0), V1 = V0
	),
	load_variant(V1, Mod),
	(   predicate_property(Mod:option_variant(_,_), _) ->
	    Mod:option_variant(Opts, Level), V = V0/Opts
	;   V = V0, Level = hard
	).

% Betlti az F varinsnak megfelel fjlt s Name-et egyesti az F
% nevvel.
load_variant(F, Name) :-
	variant_name_file(F, Name, Dir),
	working_directory(Old, Dir),
	ensure_loaded(Name:domino),
	working_directory(_, Old).

% A V varinsbeli Goal clt futtatja, ha tllpn az idkorltot, akkor
% ezt kirja s meghisul.
fail_if_time_out(Goal, V) :-
	bb_get(user:time_out, TimeOut),
	(   windows -> Goal, Res = success
	;   time_out(Goal, TimeOut, Res)
	),
	(   Res = success -> true
	;   format('~| ~`.t variant ~w timed out ~`.t~40+~n', [V]),
	    fail
	).

% Az M mtrix ltal adott feladatra a V varins
% a Sols megoldslistt adja.
domino_sols(V, M, Sols) :-
	findall(S, call_domino_variant(V, M, S), Sols).

% Sols1 s Sols2 ugyanazon megoldsok tmr listja.
eq_sols(Sols1, Sols2) :-
	ground(Sols1),
	ground(Sols2),
	same_length(Sols1, Sols2),
	sort(Sols1, SSols),
	same_length(Sols1, SSols),
	sort(Sols2, SSols).

% variant_name_file(V, N, D): a V megolds-vltozat neve N s a D 
% knyvtr tartalmazza. 
variant_name_file(V0, N, D) :-
	(   V0 = N/_ -> true
	;   N = V0
	),
	(   variant(N, D) -> true
	;   atom(N), D = N
	).

% run2(T, V, S): az V vltozatot futtatja a T ltal megadott tesztadatokra.
% Az S csatornba kirja a futsrl szl napl-bejegyzst.
run2(T, V, S) :-
	run_one(T, V, Cnts),
	writeq(S, results(V,T,Cnts)),
	write(S, '.\n'),
	flush_output(S),
	fail.
run2(_, _, _).

% run_one(T, V, Cnts): megoldja az T ltal megadott feladat
% V vltozatt. Cnts az ekzben sszegyjttt szmllk listja.
run_one(T, V, Cnts) :-
	pr(T, M),
	bb_put(test_case_counters, [sfilts=0,tfilts=0,msec=0,sols=0]),
	fd_statistics(backtracks, _),
	call_domino(V, M, Msec, Result),
	count_fd_btracks(btrks),
	incr_counter(msec, Msec),
	(   Result = success -> true
	;   incr_counter(time_out, 1)
	),
	bb_get(test_case_counters, Cnts),
	show_results(T, V, Cnts),
	incr_counters([tests=1|Cnts], total_counters).

% Kiirja a Name teszt-nevet V varinst s a Cnts0 listabeli szmllkat. 
show_results(Name, V, Cnts0) :-
	format('~w:~t~12|', [Name]),
	(   select(time_out=_, Cnts0, Cnts1) ->
	    TimeOut = ' TIME', Rel = '>='
	;   TimeOut = '     ',
	    Cnts1 = Cnts0, Rel = ' ='
	),
	select(sfilts=Succ, Cnts1, Cnts2),
	select(tfilts=Total, Cnts2, Cnts3),
	(   Total  > 0 -> FiltPerc is integer(Succ*100/Total+0.5),
	    append(Cnts3, [filt=Total,'filt%'=FiltPerc], Cnts)
	;   Cnts = Cnts3
	),
	show_counts(Cnts, '', Rel),
	format('~a ', [TimeOut]),
	current_output(S), line_position(S, Pos), N is max(1,88-Pos),
	format('~*c~w~n', [N,0' ,V]),
	flush_output.

% show_counts(Cnts, Sep, Rel): kiirja a Cnts listabeli szmllkat, mindegyiket
% a Sep szepartorral prefixlva s a szmllk neve s rtke kz a Rel atomot
% rva.
show_counts([], _Sep, _Rel).
show_counts([Name=V|Cnts], Sep, Rel) :-
	atom_length(Name, Len),
	field_width(Name, Width), Tab is Len+Width,
	format('~a~|~w~a~t~d~*+', [Sep,Name,Rel,V,Tab]),
	show_counts(Cnts, ', ', Rel).

% field_width(N, Sz): az N nev szmllt egy Sz szlessg mezbe kell kirni.
field_width(sols, W) :- !, W = 6.
field_width('filt%', W) :- !, W = 5.
field_width(_, 9).

% pr(T, M): A T feladat az M mtrixbl ll.
% A feladatot megadhatjuk '/'/2 strukturaba csomagolva, ekkor a
% struktra els argumentuma az aktulis knyvtr egy alknyvtra, a msodik
% egy az ebben az alknyvtrban tallhat test*d.txt fjlnv a * helyn
% elfordul szma.  Ha ez a struktra nem tmr, akkor az sszes ezzel
% illeszked feladatot megoldja.
% Egyebkent egy atomnak kell lennie, ami egy fajl neve,
% amely egy feladat leirasat tartalmazza.
pr(T, M) :-
	atom(T), !, domino_in(T, M).
pr(T, M) :- !,
	test(T, N), domino_in(N, M).

% Az M mtrix ltal megadott feladat V vltozatnak megoldsa Msec
% ezredmsodpercig tart.  Res = time_out ha idtllps trtnt, egybknt
% Res = success.
call_domino(V, M, Msec, Res) :-
	bb_get(user:time_out, TimeOut),
	statistics(runtime, [Msec0,_]),
	(   windows -> call_domino(V, M), Res = success
	;   time_out(call_domino(V, M), TimeOut, Res)
	),
	statistics(runtime, [Msec1,_]), Msec is Msec1-Msec0.


% Megoldja az M mtrix ltal megadott feladatot a V varinssal,
% kzben szmolja a megoldsokat.
call_domino(V, M) :-
	open_null_stream(S),
	current_output(Old), set_output(S),
	prolog_flag(user_output, UO, S),
	(   call_cleanup( call_count_fail(call_domino_variant(V, M, _Sol)),
			  (set_output(Old),
			   prolog_flag(user_output, _, UO))
			)
	;   true
	).

% call_domino_variant(V, Mx, Sol): meghvja a V varinshoz tartoz
% domino/2 ill. domino/3 eljrst.
call_domino_variant(Mod/Opts, Mx, Sol) :-
	!, Mod:domino(Mx, Sol, Opts).
call_domino_variant(Mod, Mx, Sol) :-
	Mod:domino(Mx, Sol).

% A sols vltozt megnveli Goal megoldsainak szmval, majd meghisul.
% (Ehhez persze Goal-t futtatja.)
call_count_fail(Goal) :-
	Goal, incr_counter(sols, 1), fail.

% F a Dir knyvtrbeli, Num sorszm feladatot tartalmaz fjl.
test(Dir/Num, File) :-
	directory_files(., Dirlist0),
	sort(Dirlist0, Dirlist),
	member(Dir, Dirlist),
	file_exists(Dir, [read,search,exists]),
	file_property(Dir, type(directory)),
        directory_files(Dir, Filelist0),
	sort(Filelist0, Filelist),
        member(File0, Filelist),
        atom_concat(test, FEnd, File0),
        atom_concat(NumA, 'd.txt', FEnd),
	atom_codes(NumA, NumCs),
	name(Num, NumCs), integer(Num),
        atom_concat(Dir, '/', Dir1),
        atom_concat(Dir1, File0, File),
	file_exists(File, [read]).

% not_above(L0, L): L egy L0-nal nem nehezebb szint.
not_above(L,    base) :- L \= base.
not_above(diff, easy).
not_above(hard, easy).
not_above(hard, diff).
not_above(L, L).


% Szmllk

% A legutbbi fd_statistics(backtracks, ...) lekrdezs ta eltelt fd
% visszalpsek szmval megnveli a Cnt vltoz rtkt.
count_fd_btracks(Cnt) :-
	fd_statistics(backtracks, B),
	incr_counter(Cnt, B).

% A Cnt szmll rtkt Incr-el megnveli.
incr_counter(Cnt, Incr) :-
	incr_counter(test_case_counters, Cnt, Incr).

% A Group csoportbeli Cnt szmll rtkt Incr-el megnveli.
incr_counter(Group, Cnt, Incr) :-
	bb_get(Group, Map0),
	incr_map(Map0, Cnt, Incr, Map),
	bb_put(Group, Map).

% incr_counters(Cnts, Group): a Cnts lista Nv=rtk prokbl ll, ahol minden
% Nv egy egy Group csoportbeli szmll.  Minden ilyen szmll rtkt rtkkel
% megnveli.
incr_counters([Cnt=Incr|Cnts], Group) :-
	incr_counter(Group, Cnt, Incr),
	incr_counters(Cnts, Group).
incr_counters([], _).

% incr_map(Map0, Cnt, Incr, Map): Map0 s Map listk Nv=rtk prokbl llnak.
% Map olyan, mint Map0 csak benne a Cnt prja Incr-el nagyobb, mint Map0-ban, vagy
% ha Map0 nem tartalmaz ilyen prt, akkor benne szerepel a Cnt=Incr pr.
incr_map([], Cnt, Incr, [Cnt=Incr]).
incr_map([Cnt=V0|Map0], Cnt, Incr, Map) :-
	!, V is V0+Incr,
	Map = [Cnt=V|Map0].
incr_map([CntV|Map0], Cnt, Incr, [CntV|Map]) :-
	incr_map(Map0, Cnt, Incr, Map).

% --------

% windows: sikerl, ha windows felett futunk.
:- dynamic windows/0.

% Ha mg nem tudjuk, hogy windows felett futunk, s kiderl, hogy igen,
% akkor ezt megjegyezzuk.
look_for_windows :-
	\+ windows,
	prolog_flag(version, V),
	sub_atom(V, _, _, _, win32), !,
	assert(windows).
look_for_windows.

:- look_for_windows.


% --------


% filter_vars(Vs, Ints, FVs): A Vs vltozlista Ints szrsi halmaz
% szerint megszrt vltozata az FVs.
filter_vars(Vs, Ints, FVs) :-
	count_fd_btracks(btrks),
	call_cleanup(filter_vars1(Ints, Vs, FVs, 0),
		     count_fd_btracks(sfilts)).

% filter_vars1(Ints, Vs0, FVs, Tot): A Vs0 vltozlista Ints szrsi halmaz
% szerint megszrt vltozata az FVs, ekzben Tot szm szrsi kisrlet
% trtnt.
filter_vars1([], Vs, Vs, Tot) :-
	incr_counter(tfilts, Tot).
filter_vars1([I|Is], Vs0, FVs, Tot0) :-
	filter_vars2(Vs0, I, Vs1, Tot0, Tot1),
	filter_vars1(Is, Vs1, FVs, Tot1).

% filter_vars2(Vs, I, FVs, Tot0, Tot): A Vs0 vltozlista I rtkkel
% megszrt vltozata az FVs, ekzben Tot-Tot0 szm szrsi kisrlet
% trtnt.
filter_vars2([], _I, [], Tot, Tot).
filter_vars2([V|Vs], I, FVs, Tot0, Tot) :-
        integer(V), !,
        filter_vars2(Vs, I, FVs, Tot0, Tot).
filter_vars2([V|Vs], I, [V|FVs], Tot0, Tot) :-
	Tot1 is Tot0+1,
	(   \+ V = I -> filter_value(V, I, Tot1)
        ;   true
        ),
        filter_vars2(Vs, I, FVs, Tot1, Tot).

% filter_value(V, I, Tot): A V vltoz nem egyenl az I egsszel. Eddig Tot
% szrsi kisrlet trtnt. Ha meghisul, akkor a Tot rtkkel nveli a
% szrsi kisrlek szmlljt s mdostja a visszalpsek s sikeres
% szrsek szmlljt. 
filter_value(V, I, _Tot) :-
	V #\= I, !.
filter_value(_V, _I, Tot) :-
	incr_counter(tfilts, Tot),
	incr_counter(btrks, 1),
	incr_counter(sfilts, -1),
	fail.
